/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { ConfigurationLinkRel, ConfigurationResource, ConfigurationService } from '@admin-client/configuration-shared';
import { ROUTES } from '@admin-client/shared';
import { KeycloakTokenService } from '@admin/keycloak-shared';
import { ApiRootLinkRel, ApiRootResource, ApiRootService } from '@alfa-client/api-root-shared';
import { BuildInfoComponent } from '@alfa-client/common';
import { createEmptyStateResource, createStateResource, HasLinkPipe } from '@alfa-client/tech-shared';
import { existsAsHtmlElement, getElementComponentFromFixtureByCss, Mock, mock, notExistsAsHtmlElement, } from '@alfa-client/test-utils';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { ActivatedRoute, Router, RouterOutlet } from '@angular/router';
import { AuthenticationService } from '@authentication';
import { AdminLogoIconComponent, MailboxIconComponent, NavbarComponent, NavItemComponent, OrgaUnitIconComponent, UsersIconComponent, } from '@ods/system';
import { createConfigurationResource } from 'libs/admin/configuration-shared/test/configuration';
import { MenuContainerComponent } from 'libs/admin/configuration/src/lib/menu-container/menu-container.component';
import { createApiRootResource } from 'libs/api-root-shared/test/api-root';
import { getDataTestIdOf } from 'libs/tech-shared/test/data-test';
import { MockComponent, MockDirective } from 'ng-mocks';
import { of, Subscription } from 'rxjs';
import { UserProfileButtonContainerComponent } from '../../../../libs/admin/user-profile/src/lib/user-menu/user-profile.button-container.component';
import { UnavailablePageComponent } from '../pages/unavailable/unavailable-page/unavailable-page.component';
import { AppComponent } from './app.component';

describe('AppComponent', () => {
  let component: AppComponent;
  let fixture: ComponentFixture<AppComponent>;

  const adminHeaderSelector: string = getDataTestIdOf('admin-header');
  const buildInfoSelector: string = getDataTestIdOf('build-info');
  const userProfileButtonSelector: string = getDataTestIdOf('user-profile-button');
  const usersRolesNavigationSelector: string = getDataTestIdOf('users-roles-navigation');

  const organisationsEinheitenNavigationSelector: string = getDataTestIdOf('organisations-einheiten-navigation');
  const routerOutletSelector: string = getDataTestIdOf('router-outlet');
  const menuContainer: string = getDataTestIdOf('menu-container');

  let authenticationService: Mock<AuthenticationService>;
  let router: Mock<Router>;
  let route: Mock<ActivatedRoute>;
  let apiRootService: Mock<ApiRootService>;
  let configurationService: Mock<ConfigurationService>;
  let keycloakTokenService: Mock<KeycloakTokenService>;

  beforeEach(async () => {
    authenticationService = {
      ...mock(AuthenticationService),
      login: jest.fn().mockResolvedValue(Promise.resolve()),
    };
    router = mock(Router);
    route = {
      ...mock(ActivatedRoute),
      snapshot: {
        queryParams: {
          iss: 'some-iss',
          state: 'some-state',
          session_state: 'some-session-state',
          code: 'some-code',
        },
      } as any,
    };
    apiRootService = mock(ApiRootService);
    configurationService = mock(ConfigurationService);
    keycloakTokenService = mock(KeycloakTokenService);

    await TestBed.configureTestingModule({
      imports: [HasLinkPipe],
      declarations: [
        AppComponent,
        MockComponent(AdminLogoIconComponent),
        MockComponent(OrgaUnitIconComponent),
        MockComponent(UsersIconComponent),
        MockComponent(MailboxIconComponent),
        MockComponent(UserProfileButtonContainerComponent),
        MockComponent(UnavailablePageComponent),
        MockComponent(NavbarComponent),
        MockComponent(NavItemComponent),
        MockComponent(BuildInfoComponent),
        MockComponent(MenuContainerComponent),
        MockDirective(RouterOutlet),
      ],
      providers: [
        {
          provide: AuthenticationService,
          useValue: authenticationService,
        },
        {
          provide: ApiRootService,
          useValue: apiRootService,
        },
        {
          provide: ConfigurationService,
          useValue: configurationService,
        },
        {
          provide: Router,
          useValue: router,
        },
        {
          provide: ActivatedRoute,
          useValue: route,
        },
        {
          provide: KeycloakTokenService,
          useValue: keycloakTokenService,
        },
      ],
    }).compileComponents();
  });

  beforeEach(() => {
    fixture = TestBed.createComponent(AppComponent);
    component = fixture.componentInstance;
  });

  describe('component', () => {
    it(`should have as title 'admin'`, () => {
      const appTitle: string = fixture.componentInstance.title;

      expect(appTitle).toEqual('admin');
    });

    describe('ngOnInit', () => {
      it('should call authService login', () => {
        component.ngOnInit();

        expect(authenticationService.login).toHaveBeenCalled();
      });

      it('should call doAfterLoggedIn only once', async () => {
        component._doAfterLoggedIn = jest.fn();

        component.ngOnInit();
        await fixture.whenStable();

        expect(component._doAfterLoggedIn).toHaveBeenCalledTimes(1);
      });
    });

    describe('do after logged in', () => {
      beforeEach(() => {
        component._evaluateInitialNavigation = jest.fn();
      });

      it('should call apiRootService to getApiRoot', () => {
        component._doAfterLoggedIn();

        expect(apiRootService.getApiRoot).toHaveBeenCalled();
      });

      it('should call keycloak token service to register token provider', () => {
        component._doAfterLoggedIn();

        expect(keycloakTokenService.registerAccessTokenProvider).toHaveBeenCalled();
      });

      it('should call evaluateInitialNavigation', () => {
        component._doAfterLoggedIn();

        expect(component._evaluateInitialNavigation).toHaveBeenCalled();
      });
    });

    describe('evaluate initial navigation', () => {
      beforeEach(() => {
        component._evaluateNavigationByApiRoot = jest.fn();
        component._subscribeApiRootForEvaluation = jest.fn();
      });

      it('should call router navigate', () => {
        window.location.pathname = '/path';
        (router as any).url = '/path';

        component._evaluateInitialNavigation();

        expect(router.navigate).toHaveBeenCalledWith([window.location.pathname]);
      });

      it('should call subscribe api root evaluation if url starts with /?state', () => {
        (router as any).url = '/?state=some-state';

        component._evaluateInitialNavigation();

        expect(component._subscribeApiRootForEvaluation).toHaveBeenCalled();
      });
    });

    describe('subscribeApiRootForEvaluation', () => {
      const apiRootResource: ApiRootResource = createApiRootResource();

      beforeEach(() => {
        component.apiRootStateResource$ = of(createStateResource(apiRootResource));
        component._evaluateNavigationByApiRoot = jest.fn();
      });

      it('should set apiRootSubscription', () => {
        component._subscribeApiRootForEvaluation();

        expect(component.apiRootSubscription).toBeDefined();
      });

      it('should call evaluate navigation by apiRoot', () => {
        component._subscribeApiRootForEvaluation();

        expect(component._evaluateNavigationByApiRoot).toHaveBeenCalledWith(apiRootResource);
      });

      it('should NOT call evaluate navigation by apiRoot if resource is loading', () => {
        component.apiRootStateResource$ = of(createStateResource(apiRootResource, true));

        component._subscribeApiRootForEvaluation();

        expect(component._evaluateNavigationByApiRoot).not.toHaveBeenCalled();
      });
    });

    describe('evaluate navigation api root', () => {
      it('should evaluate navigation by configuration if link exists', () => {
        component._evaluateNavigationByConfiguration = jest.fn();
        const apiRootResource: ApiRootResource = createApiRootResource([ApiRootLinkRel.CONFIGURATION]);

        component._evaluateNavigationByApiRoot(apiRootResource);

        expect(component._evaluateNavigationByConfiguration).toHaveBeenCalled();
      });

      it('should navigate by api root if link is missing', () => {
        component._navigateByApiRoot = jest.fn();
        const apiRootResource: ApiRootResource = createApiRootResource();

        component._evaluateNavigationByApiRoot(apiRootResource);

        expect(component._navigateByApiRoot).toHaveBeenCalledWith(apiRootResource);
      });
    });

    describe('evaluate navigation by configuration', () => {
      const configurationResource: ConfigurationResource = createConfigurationResource();

      beforeEach(() => {
        configurationService.get.mockReturnValue(of(createStateResource(configurationResource)));
        component._navigateByConfiguration = jest.fn();
      });

      it('should call configuration service to get resource', () => {
        component._evaluateNavigationByConfiguration();

        expect(configurationService.get).toHaveBeenCalled();
      });

      it('should call navigate by configuration', () => {
        component._evaluateNavigationByConfiguration();

        expect(component._navigateByConfiguration).toHaveBeenCalledWith(configurationResource);
      });

      it('should NOT call navigate by configuration if resource is loading', () => {
        configurationService.get.mockReturnValue(of(createEmptyStateResource<ConfigurationResource>(true)));

        component._evaluateNavigationByConfiguration();

        expect(component._navigateByConfiguration).not.toHaveBeenCalled();
      });
    });

    describe('navigate by configuration', () => {
      beforeEach(() => {
        component._unsubscribe = jest.fn();
      });

      it('should navigate to postfach if settings link exists', () => {
        component._navigateByConfiguration(createConfigurationResource([ConfigurationLinkRel.SETTING]));

        expect(router.navigate).toHaveBeenCalledWith(['/postfach']);
      });

      it('should navigate to statistik if aggregation mapping link exists', () => {
        component._navigateByConfiguration(createConfigurationResource([ConfigurationLinkRel.AGGREGATION_MAPPINGS]));

        expect(router.navigate).toHaveBeenCalledWith(['/statistik']);
      });

      it('should navigate to unavailable page if no link exists', () => {
        component._navigateByConfiguration(createConfigurationResource());

        expect(router.navigate).toHaveBeenCalledWith(['/unavailable']);
      });

      it('should call unsubscribe', () => {
        component._navigateByConfiguration(createConfigurationResource());

        expect(component._unsubscribe).toHaveBeenCalled();
      });
    });

    describe('navigate by api root', () => {
      beforeEach(() => {
        component._unsubscribe = jest.fn();
      });

      it('should navigate to benutzer und rollen if users link exists', () => {
        const apiRootResource: ApiRootResource = createApiRootResource([ApiRootLinkRel.USERS]);

        component._navigateByApiRoot(apiRootResource);

        expect(router.navigate).toHaveBeenCalledWith([`/${ROUTES.BENUTZER}`]);
      });

      it('should navigate to unavailable page if no link exists', () => {
        component._navigateByApiRoot(createApiRootResource());

        expect(router.navigate).toHaveBeenCalledWith([`/${ROUTES.UNAVAILABLE}`]);
      });

      it('should call unsubscribe', () => {
        component._navigateByApiRoot(createApiRootResource());

        expect(component._unsubscribe).toHaveBeenCalled();
      });
    });

    describe('unsubscribe', () => {
      describe('apiRoot subscription', () => {
        it('should unsubscribe if exists', () => {
          component.apiRootSubscription = <any>mock(Subscription);
          component.apiRootSubscription.unsubscribe = jest.fn();

          component._unsubscribe();

          expect(component.apiRootSubscription.unsubscribe).toHaveBeenCalled();
        });

        it('should do nothing if not exists', () => {
          component.apiRootSubscription = undefined;
          const unsubscribeSpy: jest.SpyInstance = jest.spyOn(Subscription.prototype, 'unsubscribe');

          component._unsubscribe();

          expect(unsubscribeSpy).not.toHaveBeenCalled();
          unsubscribeSpy.mockRestore();
        });
      });

      describe('configuration subscription', () => {
        it('should unsubscribe if exists', () => {
          component.configurationSubscription = <any>mock(Subscription);
          component.configurationSubscription.unsubscribe = jest.fn();

          component._unsubscribe();

          expect(component.configurationSubscription.unsubscribe).toHaveBeenCalled();
        });

        it('should do nothing if not exists', () => {
          component.apiRootSubscription = undefined;
          const unsubscribeSpy: jest.SpyInstance = jest.spyOn(Subscription.prototype, 'unsubscribe');

          component._unsubscribe();

          expect(unsubscribeSpy).not.toHaveBeenCalled();
          unsubscribeSpy.mockRestore();
        });
      });
    });
  });

  describe('template', () => {
    it('show not show header if apiRoot is not loaded', () => {
      component.apiRootStateResource$ = of(createEmptyStateResource<ApiRootResource>());

      notExistsAsHtmlElement(fixture, adminHeaderSelector);
    });

    describe('user profile button', () => {
      beforeEach(() => {
        component.apiRootStateResource$ = of(createStateResource(createApiRootResource()));
      });

      it('should show if apiRoot exists', () => {
        fixture.detectChanges();

        existsAsHtmlElement(fixture, userProfileButtonSelector);
      });
    });

    describe('navigation', () => {
      describe('user and roles', () => {
        it('should have inputs', () => {
          component.apiRootStateResource$ = of(
            createStateResource(createApiRootResource([ApiRootLinkRel.CONFIGURATION, ApiRootLinkRel.USERS])),
          );

          fixture.detectChanges();
          const naviItemComponent: NavItemComponent = getElementComponentFromFixtureByCss<NavItemComponent>(
            fixture,
            usersRolesNavigationSelector,
          );

          expect(naviItemComponent.path).toEqual(ROUTES.BENUTZER);
        });

        it('should show if users link is present', () => {
          component.apiRootStateResource$ = of(
            createStateResource(createApiRootResource([ApiRootLinkRel.CONFIGURATION, ApiRootLinkRel.USERS])),
          );

          fixture.detectChanges();

          existsAsHtmlElement(fixture, usersRolesNavigationSelector);
        });

        it('should hide if link is missing', () => {
          component.apiRootStateResource$ = of(createStateResource(createApiRootResource([ApiRootLinkRel.CONFIGURATION])));

          fixture.detectChanges();

          notExistsAsHtmlElement(fixture, usersRolesNavigationSelector);
        });
      });

      describe('organisationEinheiten', () => {
        it('should have inputs', () => {
          component.apiRootStateResource$ = of(
            createStateResource(createApiRootResource([ApiRootLinkRel.ORGANISATIONS_EINHEIT, ApiRootLinkRel.CONFIGURATION])),
          );

          fixture.detectChanges();
          const naviItemComponent: NavItemComponent = getElementComponentFromFixtureByCss<NavItemComponent>(
            fixture,
            organisationsEinheitenNavigationSelector,
          );

          expect(naviItemComponent.path).toEqual(ROUTES.ORGANISATIONSEINHEITEN);
        });

        it('should show if link in apiRoot exists', () => {
          component.apiRootStateResource$ = of(
            createStateResource(createApiRootResource([ApiRootLinkRel.ORGANISATIONS_EINHEIT, ApiRootLinkRel.CONFIGURATION])),
          );
          fixture.detectChanges();

          existsAsHtmlElement(fixture, organisationsEinheitenNavigationSelector);
        });

        it('should hide if link in apiRoot does not exist', () => {
          fixture.detectChanges();

          notExistsAsHtmlElement(fixture, organisationsEinheitenNavigationSelector);
        });
      });

      describe('menu', () => {
        it('should show if link exists', () => {
          component.apiRootStateResource$ = of(createStateResource(createApiRootResource([ApiRootLinkRel.CONFIGURATION])));

          fixture.detectChanges();

          existsAsHtmlElement(fixture, menuContainer);
        });

        it('should hide if link is missing', () => {
          component.apiRootStateResource$ = of(createStateResource(createApiRootResource()));

          fixture.detectChanges();

          notExistsAsHtmlElement(fixture, menuContainer);
        });
      });
    });

    describe('build version', () => {
      it('should not be rendered if api root not loaded', () => {
        notExistsAsHtmlElement(fixture, buildInfoSelector);
      });

      it('should show after apiRoot loaded', () => {
        component.apiRootStateResource$ = of(createStateResource(createApiRootResource()));
        fixture.detectChanges();

        existsAsHtmlElement(fixture, buildInfoSelector);
      });
    });

    describe('router outlet', () => {
      beforeEach(() => {});

      it('should exist if configuration resource available', () => {
        component.apiRootStateResource$ = of(createStateResource(createApiRootResource([ApiRootLinkRel.CONFIGURATION])));

        fixture.detectChanges();

        existsAsHtmlElement(fixture, routerOutletSelector);
      });

      it('should not exist if configuration resource not available', () => {
        component.apiRootStateResource$ = of(createStateResource(createApiRootResource()));

        fixture.detectChanges();

        notExistsAsHtmlElement(fixture, routerOutletSelector);
      });
    });
  });
});
